import torch
from typing import Optional, Dict
from transformers.cache_utils import Cache

class LLmComputationState:
    state: torch.Tensor
    position_embeddings: Optional[torch.Tensor]
    position_embeddings_local: Optional[torch.Tensor]
    position_embeddings_global: Optional[torch.Tensor]
    position_ids: torch.Tensor
    cache_position: torch.Tensor
    causal_mask: Dict
    past_key_values: Cache

    def __init__(self):
        self.state = None
        self.position_embeddings = None
        self.position_embeddings_local = None
        self.position_embeddings_global = None
        self.position_ids = None
        self.cache_position = None
        self.causal_mask = { }
        self.past_key_values = None


import gc
import torch
from typing import List, Dict

from safetensors import safe_open
from transformers.configuration_utils import PretrainedConfig
from llm_layer_collector.auto.auto_layer import AutoDecoderLayer

def files_to_load_for_layer(
        layer_prefix: str,
        layer_file_cache: dict,
    ) -> List[str]:
    files_to_load = []
    for key in layer_file_cache.keys():
        if key.startswith(layer_prefix) and layer_file_cache[key] not in files_to_load:
            files_to_load.append(layer_file_cache[key])
    if len(files_to_load) == 0:
        raise Exception("Could not find layer data for layer prefix " + layer_prefix)
    return files_to_load

def files_to_load_for_layers(
        start_layer: int,
        end_layer: int,
        layer_prefix: str,
        layer_file_cache: dict
    ) -> List[str]:
    files_to_load = []
    for i in range(start_layer, end_layer+1):
        for f in files_to_load_for_layer(f'{layer_prefix}{i}.', layer_file_cache):
            if f not in files_to_load:
                files_to_load.append(f)
    return files_to_load

def get_shard_data(
        start_layer: int,
        end_layer: int,
        device: str,
        model_dir: str,
        layer_prefix: str,
        layer_file_cache: Dict[str, str],
        dtype: str
    ) -> Dict[str, torch.Tensor]:
    prefixes = [f'{layer_prefix}{i}.' for i in range(start_layer, end_layer+1)]
    shard_data = { }
    for file_path in files_to_load_for_layers(start_layer, end_layer, layer_prefix, layer_file_cache):
        full_path = f'{model_dir}/{file_path}'
        shard: dict = safe_open(full_path, framework='pt', device=device)
        for key in shard.keys():
            for prefix in prefixes:
                if key.startswith(prefix):
                    shard_data[key] = shard.get_tensor(key).detach().to(dtype)
        del shard
        gc.collect()
    
    return shard_data

def load_layer(
        config: PretrainedConfig, 
        idx: int, 
        shard_data: Dict,
        layer_prefix: str,
        device: str,
        dtype: str
    ) -> AutoDecoderLayer:
    torch.set_default_device('meta')
    lyr = AutoDecoderLayer(config, idx)
    torch.set_default_device(device)
    lyr = lyr.to_empty(device=device)
    for key in shard_data.keys():
        if key.startswith(f'{layer_prefix}{idx}.'):
            module_name = key.replace(f'{layer_prefix}{idx}.', '').replace('.weight', '')
            module = lyr.get_submodule(module_name)
            module.load_state_dict({ "weight": shard_data[key] })

    return lyr.to(dtype)

def load_layers(
        start_layer: int, 
        end_layer: int, 
        layer_prefix: str,
        layer_file_cache: Dict[str, str],
        config: PretrainedConfig,
        model_dir: str,
        device: str,
        dtype: str
    ) -> List[AutoDecoderLayer]:
    torch.set_default_device(device)
    shard_data = get_shard_data(start_layer, end_layer, device, model_dir, layer_prefix, layer_file_cache, dtype)
    layers = []
    for i in range(start_layer, end_layer+1):
        layers.append(load_layer(config, i, shard_data, layer_prefix, device, dtype))

    torch.set_default_device('cpu')
    return layers



import copy
import torch
from typing import Optional
from transformers.configuration_utils import PretrainedConfig
from transformers.masking_utils import create_causal_mask, create_sliding_window_causal_mask
from transformers.modeling_attn_mask_utils import AttentionMaskConverter
from transformers.cache_utils import DynamicCache

from llm_layer_collector.auto.auto_rotary import AutoRotaryEmbedding
from llm_layer_collector.state_obj import LLmComputationState

def compute_embedding(
        input_embedder: torch.nn.Embedding,
        input_ids: torch.Tensor,
        config: PretrainedConfig,
        state: Optional[LLmComputationState] = None
    ) -> LLmComputationState:
    device = input_embedder.weight.device
    embedded_input = input_embedder(input_ids.to(device))
    if state is None:
        state = LLmComputationState()
    
    state.state = embedded_input

    converter = AttentionMaskConverter(is_causal=True)
    L = embedded_input.size()[1]
    attention_mask = converter.to_causal_4d(
        batch_size=1,
        query_length=L,
        key_value_length=L,
        dtype=embedded_input.dtype,
        device=embedded_input.device
    )
    
    state.cache_position = torch.arange(
        0, end=embedded_input.size(1), device=device
    )
    
    state.position_ids = state.cache_position.unsqueeze(0)

    mask_kwargs = {
        "config": config,
        "input_embeds": embedded_input.detach(),
        "attention_mask": attention_mask,
        "cache_position": state.cache_position,
        "past_key_values": state.past_key_values,
        "position_ids": state.position_ids
    }
    
    state.causal_mask["full_attention"] = create_causal_mask(**mask_kwargs)

    try:
        if "sliding_attention" in config.layer_types or config.sliding_window is not None:
            state.causal_mask["sliding_attention"] = create_sliding_window_causal_mask(**mask_kwargs)
    except AttributeError:
        pass

    if config.model_type == 'gemma3_text':
        state.position_embeddings_global = AutoRotaryEmbedding(config)(embedded_input.detach(), state.position_ids)
        configCopy = copy.deepcopy(config)
        configCopy.rope_theta = configCopy.rope_local_base_freq
        configCopy.rope_scaling = {"rope_type": "default"}
        
        state.position_embeddings_local = AutoRotaryEmbedding(configCopy)(embedded_input.detach(), state.position_ids)
    else:
        state.position_embeddings = AutoRotaryEmbedding(config)(embedded_input.detach(), state.position_ids)
    
    return state

def compute_head(
        head: torch.nn.Linear,
        state: torch.Tensor,
        topk: int = 1
    ) -> torch.Tensor:
    state = head(state[:, -1, :])
    probs = torch.softmax(state, dim=-1)
    return torch.topk(probs, topk).indices



import re
import os
from typing import List

from safetensors import safe_open

def get_shard_files(shard_pattern: str, model_dir: str) -> List[str]:
    if 'model.safetensors' in os.listdir(model_dir):
        return ['model.safetensors']
    
    multiple_pattern = re.compile(shard_pattern)
    shard_files = [f for f in os.listdir(model_dir) if multiple_pattern.match(f)]
    if not shard_files:
        raise Exception("No Shard files in specified directory " + model_dir)

    shard_files.sort()
    return shard_files

def build_cache_data(
        model_dir: str,
        shard_pattern: str,
        device: str
    ):
    layer_files = { }
    for file in get_shard_files(shard_pattern, model_dir):
        full_path = os.path.join(model_dir, file)
        shard: dict = safe_open(full_path, framework='pt', device=device)
        for key in shard.keys():
            layer_files[key] = file
        del shard

    return layer_files


import os

import torch
from safetensors import safe_open

def load_shard_tensor(
        layer_file_cache: dict, 
        model_dir: str,
        layer_name: str, 
        device: str,
        dtype: torch.dtype
    ) -> torch.Tensor:
    if layer_name not in layer_file_cache:
        raise ValueError(f'Could not find layer file for layer {layer_name}')
    file = layer_file_cache[layer_name]
    shard: dict = safe_open(os.path.join(model_dir, file), framework='pt', device=device)
    return shard.get_tensor(layer_name).to(dtype)


import os
import sys
import json
import unittest
import subprocess
from time import time
from pathlib import Path

import torch
from torch import tensor
from transformers import AutoTokenizer

sys.path.append(os.path.join(os.path.dirname(__file__), '../src'))

from llm_layer_collector.layer_collector import LlmLayerCollector
from llm_layer_collector.compute import compute_embedding, compute_head
from llm_layer_collector.cache import get_shard_files
from llm_layer_collector.helpers import load_shard_tensor
from llm_layer_collector.load_layer import files_to_load_for_layer

PROMPT = "The quick brown fox jumps over the "

def clone_model(model_id: str, model_dir: str):
    repo_url = f"https://huggingface.co/{model_id}"

    if not os.path.exists(model_dir):
        Path(model_dir).mkdir(parents=True)
    subprocess.run(["git", "clone", repo_url, model_dir])
    subprocess.run(["git", "lfs", "install"], cwd=model_dir, check=True)
    subprocess.run(["git", "lfs", "pull"], cwd=model_dir, check=True)

def get_cache_file(model_id: str):
    model_str = model_id.replace('/', '_')
    return f"data/{model_str}.json"

def get_model_dir(model_id: str):
    return f"models/{model_id}"

def ensure_model(model_id: str):
    model_dir = get_model_dir(model_id)
    if os.path.exists(model_dir):
        return
    clone_model(model_id, model_dir)

def test_cache(tst: unittest.TestCase, model_dir: str, cache_file: str, num_keys: int):
    collector = LlmLayerCollector(model_dir, cache_file)
    tst.assertEqual(len(collector.layer_files.keys()), num_keys)
    tst.assertTrue(os.path.exists(cache_file))
    tst.assertTrue(os.path.exists(collector.cache_file))
    with open(cache_file, 'r') as f:
        cache = json.load(f)
        tst.assertEqual(len(cache.keys()), num_keys)

def test_embedding(tst: unittest.TestCase, model_dir: str, cache_file: str, state_shape, position_ids_shape, position_embeddings_shape):
    collector = LlmLayerCollector(model_dir, cache_file)
    input_embedder = collector.load_input_embedding()
    tokenizer = AutoTokenizer.from_pretrained(model_dir)
    input_ids = tokenizer(PROMPT, return_tensors='pt')['input_ids']
    state = compute_embedding(input_embedder, input_ids, collector.config)
    tst.assertEqual(state.state.shape, state_shape)
    tst.assertEqual(state.position_ids.shape, position_ids_shape)
    tst.assertEqual(state.position_embeddings[0].shape, position_embeddings_shape)
    tst.assertEqual(state.position_embeddings[1].shape, position_embeddings_shape)

def test_norm(tst: unittest.TestCase, model_dir: str, cache_file: str, norm_dim: int):
    collector = LlmLayerCollector(model_dir, cache_file)
    norm = collector.load_norm()
    norm = norm.to('cpu')
    norm = norm.to(dtype=torch.float16)
    tst.assertEqual(norm.weight.shape, (norm_dim,))

def test_head(tst: unittest.TestCase, model_dir: str, cache_file: str, head_shape):
    collector = LlmLayerCollector(model_dir, cache_file)
    head = collector.load_head()
    tst.assertEqual(head.weight.shape, head_shape)

def test_layers(tst: unittest.TestCase, model_dir: str, cache_file: str, end_layer: int):
    start_time = time()
    collector = LlmLayerCollector(model_dir, cache_file)
    layers = collector.load_layer_set(0, end_layer)
    print(f"Time: {time() - start_time:.2f}s")
    tst.assertEqual(len(layers), end_layer+1)

def test_stack(tst: unittest.TestCase, model_dir: str, cache_file: str):
    tokenizer = AutoTokenizer.from_pretrained(model_dir)
    chat = [
        {"role": "system", "content": "You are a friendly chatbot who always responds in the style of a pirate",},
        {"role": "user", "content": "How many helicopters can a human eat in one sitting?"},
    ]
    input_ids = tokenizer.apply_chat_template(chat, tokenize=True, add_generation_prompt=True, return_tensors='pt')
    original_num_tokens = input_ids.shape[1]
    num_tokens = 4
    current_token = 0
    collector = LlmLayerCollector(model_dir, cache_file)
    input_embed = collector.load_input_embedding()
    head = collector.load_head()
    norm = collector.load_norm()
    layers = collector.load_layer_set(0, collector.config.num_hidden_layers - 1) # Ensure we load all layers....
    state = None
    while current_token < num_tokens:
        state = compute_embedding(input_embed, input_ids, collector.config, state)
        for lyr in layers:
            state.state = lyr(state)
        topk = 1
        result = compute_head(head, norm(state.state), topk)
        tst.assertEqual(result.shape, (1, topk))
        token_list = input_ids.tolist()[0]
        token_list.append(result[0][0].item())
        input_ids = tensor([token_list])
        current_token += 1
        print(current_token)
        print(tokenizer.decode(input_ids[0]))
    tst.assertGreater(input_ids.shape[1], original_num_tokens)

class LlmLayerCollectorTests(unittest.TestCase):
    def test_qwen3_2B(self):
        model_id = "Qwen/Qwen3-1.7B"
        model_dir = get_model_dir(model_id)
        cache_file = get_cache_file(model_id)
        ensure_model(model_id)
        test_cache(self, model_dir, cache_file, 311)
        test_embedding(self, model_dir, cache_file, (1, 8, 2048), (1, 8), (1, 8, 128))
        test_norm(self, model_dir, cache_file, 2048)
        test_head(self, model_dir, cache_file, (151936, 2048))
        test_layers(self, model_dir, cache_file, 10)
        test_stack(self, model_dir, cache_file)

    # def test_gemma3_1B(self):
    #     model_id = "google/gemma-3-1b-it"
    #     model_dir = get_model_dir(model_id)
    #     cache_file = get_cache_file(model_id)
    #     ensure_model(model_id)

    # def test_gpt_oss_20B(self):
    #     model_id = "openai/gpt-oss-20b"
    #     model_dir = get_model_dir(model_id)
    #     cache_file = get_cache_file(model_id)
    #     ensure_model(model_id
    #     test_cache(self, model_dir, cache_file, 459)
    #     test_embedding(self, model_dir, cache_file, (1, 8, 2048), (1, 8), (1, 8, 128))
    #     test_norm(self, model_dir, cache_file, 2048)
    #     test_head(self, model_dir, cache_file, (151936, 2048))
    #     test_layers(self, model_dir, cache_file, 10)

    def test_exceptions(self):
        model_id = "Qwen/Qwen3-1.7B"
        model_dir = get_model_dir(model_id)
        cache_file = get_cache_file(model_id)
        collector = LlmLayerCollector(model_dir, cache_file)

        try:
            os.mkdir('shard_test')
            get_shard_files(collector.shard_pattern, 'shard_test')
            self.fail("Should have thrown an exception")
        except Exception:
            pass
        os.rmdir('shard_test')
        
        try:
            load_shard_tensor(collector.layer_files, collector.model_dir, 'bad_layer', 'cpu', torch.float16)
            self.fail("Should have thrown an exception")
        except ValueError:
            pass

        try:
            os.mkdir('bad_dir')
            LlmLayerCollector('bad_dir', cache_file)
            self.fail("Should have thrown an exception")
        except FileNotFoundError:
            pass

        os.rmdir('bad_dir')

        try:
            os.remove(cache_file)
            collector._read_cache()
            self.fail("Should have thrown an exception")
        except FileNotFoundError:
            pass
        
        try:
            files_to_load_for_layer('bad_key', [])
            self.fail("Should have thrown an exception")
        except Exception:
            pass

# If you want to run these tests directly from the command line:
if __name__ == '__main__':
    unittest.main()



import torch
from huggingface_hub import login
from transformers import AutoModelForCausalLM, AutoTokenizer

login()

tokenizer = AutoTokenizer.from_pretrained("meta-llama/Llama-3.2-1B-Instruct")
model = AutoModelForCausalLM.from_pretrained("meta-llama/Llama-3.2-1B-Instruct")

num_tokens = 10

input_ids = tokenizer("The quick brown fox jumps over the ")['input_ids']
for _ in range(num_tokens):
    output = model(torch.tensor([input_ids]), use_cache=False)

    probs = torch.softmax(output.logits[:, -1, :][0], dim=-1)
    top = int(torch.topk(probs, 1).indices[0])

    input_ids.append(top)
    print(tokenizer.decode(input_ids))


import os
import gc
import json
import tqdm
from typing import List, Dict, Optional

import torch
from transformers.configuration_utils import PretrainedConfig
from transformers.models.auto import AutoConfig

from llm_layer_collector.load_layer import load_layers
from llm_layer_collector.cache import build_cache_data
from llm_layer_collector.helpers import load_shard_tensor
from llm_layer_collector.auto.auto_rms import AutoRMSNorm
from llm_layer_collector.auto.auto_layer import AutoDecoderLayer

class LlmLayerCollector:
    layer_prefix: str
    norm_layer_name: str
    input_embedding_layer_name: str
    lm_head_name: str
    shard_pattern: str
    
    config: PretrainedConfig
    
    model_dir: str
    cache_file: str

    num_layers: int
    num_shards: int
    dtype: torch.dtype
    device: str
    layer_files: Dict[str, str]

    def __init__(
            self, 
            model_dir: str,
            cache_file: Optional[str] = None,
            shard_pattern: str = r'model-(\d+)-of-(\d+).safetensors',
            layer_prefix: str = 'model.layers.',
            input_embedding_layer_name: str = 'model.embed_tokens.weight',
            norm_layer_name: str = 'model.norm.weight',
            lm_head_name: str = 'lm_head.weight',
            dtype: torch.dtype = torch.float16,
            device: str = 'cpu'
        ):
        config_file_path = os.path.join(model_dir, 'config.json')
        if not os.path.exists(config_file_path):
            raise FileNotFoundError('Could not find config file ' + config_file_path)
        
        config = AutoConfig.from_pretrained(model_dir)
        self.config = config
        if "_attn_implementation" not in self.config:
            self.config._attn_implementation = "sdpa"
        self.num_layers = self.config.num_hidden_layers
        
        self.model_dir = model_dir
        
        self.lm_head_name = lm_head_name
        self.layer_prefix = layer_prefix
        self.norm_layer_name = norm_layer_name
        self.input_embedding_layer_name = input_embedding_layer_name
        self.shard_pattern = shard_pattern

        self.dtype = dtype
        self.device = device
        self.layer_files = { }
        if cache_file is None:
            raise Exception("Must provide cache file path")
        self.cache_file = cache_file

        if not os.path.exists(self.cache_file):
            self._build_cache()
        else:
            self._read_cache()

    def _read_cache(self):
        if not os.path.exists(self.cache_file):
            raise FileNotFoundError('Could not find cache file ' + self.cache_file)
        with open(self.cache_file, 'r', encoding='utf-8') as f:
            self.layer_files = json.load(f)

    # Use model.safetensors.index.json by default if it exists
    def _build_cache(self):
        self.layer_files = build_cache_data(self.model_dir, self.shard_pattern, self.device)
        if self.cache_file is not None:
            with open(self.cache_file, 'w', encoding='utf-8') as f:
                json.dump(self.layer_files, f, indent=4)

    def _load_shard_tensor(self, layer_name: str, device: str) -> torch.Tensor:
        return load_shard_tensor(self.layer_files, self.model_dir, layer_name, device, self.dtype)

    def load_input_embedding(self, device: Optional[str] = None) -> torch.nn.Embedding:
        device = self.device if device is None else device
        return torch.nn.Embedding.from_pretrained(self._load_shard_tensor(self.input_embedding_layer_name, device))
    
    def load_norm(self, device: Optional[str] = None) -> AutoRMSNorm:
        device = self.device if device is None else device
        norm = AutoRMSNorm(self.config)
        norm.weight = torch.nn.Parameter(self._load_shard_tensor(self.norm_layer_name, device))
        return norm
    
    def load_head(self, device: Optional[str] = None) -> torch.nn.Linear:
        device = self.device if device is None else device
        weight = None
        
        if self.lm_head_name is None or self.lm_head_name not in self.layer_files:
            weight = self.load_input_embedding(device).weight
        else:
            weight = self._load_shard_tensor(self.lm_head_name, device)

        head = torch.nn.Linear(weight.size()[1], weight.size()[0], device=device, dtype=self.dtype)
        head.weight = torch.nn.Parameter(weight)
        return head

    def load_layer_set(self, start_layer: int, end_layer: int, device: Optional[str] = None) -> List[AutoDecoderLayer]:
        device = self.device if device is None else device
        layers = []
        for i in tqdm.tqdm(range(start_layer, end_layer+1, 3)):
            layers.extend(load_layers(min(i, end_layer), min(i+2, end_layer), self.layer_prefix, self.layer_files, self.config, self.model_dir, device, self.dtype))
        gc.collect()
        return layers



from typing import Optional

import torch
from transformers.modeling_layers import GradientCheckpointingLayer
from transformers.configuration_utils import PretrainedConfig
from transformers.models.llama.modeling_llama import LlamaDecoderLayer
from transformers.models.qwen3.modeling_qwen3 import Qwen3DecoderLayer
from transformers.models.gemma3.modeling_gemma3 import Gemma3DecoderLayer
from transformers.models.qwen3_moe.modeling_qwen3_moe import Qwen3MoeDecoderLayer

from llm_layer_collector.state_obj import LLmComputationState

mapper = {
    "llama": LlamaDecoderLayer,
    "qwen3": Qwen3DecoderLayer,
    "gemma3_text": Gemma3DecoderLayer,
    "qwen3_moe": Qwen3MoeDecoderLayer
}

def getClass(config: PretrainedConfig) -> GradientCheckpointingLayer:
    return mapper[config.model_type]

class AutoDecoderLayer:
    def __init__(self, config: PretrainedConfig, layer_index: int):
        self.config = config
        if self.config._attn_implementation is None:
            self.config._attn_implementation = "eager"
        self.cls = getClass(self.config)(self.config, layer_index)

    def __call__(
        self, 
        state: LLmComputationState
    ) -> torch.Tensor:
        attention_type = "full_attention"
        try:
            attention_type = self.cls.attention_type
        except AttributeError:
            pass

        try:
            if self.config.sliding_window is not None:
                attention_type = "sliding_attention"
        except AttributeError:
            pass

        kwargs = {
            "hidden_states": state.state,
            "attention_mask": state.causal_mask[attention_type],
            "position_ids": state.position_ids,
            "use_cache": self.config.use_cache,
            "cache_position": state.cache_position,
        }

        if self.config.model_type == "qwen3" or self.config.model_type == "qwen3_moe":
            kwargs["past_key_values"] = state.past_key_values
        else:
            kwargs["past_key_value"] = state.past_key_values

        if self.config.model_type == "gemma3_text":
            kwargs["position_embeddings_local"] = state.position_embeddings_local
            kwargs["position_embeddings_global"] = state.position_embeddings_global
            return self.cls(**kwargs)[0]
        else:
            kwargs["position_embeddings"] = state.position_embeddings
            return self.cls(**kwargs)

    def to_empty(self, device: Optional[str]) -> 'AutoDecoderLayer':
        self.cls = self.cls.to_empty(device=device)
        return self

    def get_submodule(self, module_name: str):
        return self.cls.get_submodule(module_name)

    def to(self, device: str):
        self.cls = self.cls.to(device)
        return self